/*
 * CheckableListItem.cpp
 *
 *  Created on: 8.1.2010
 *      Author: apoi
 */

#include <QKeyEvent>
#include <QImageReader>
#include <QTimer>

#include "CheckableListItem.h"
#include "ui_CheckableListItem.h"
#include "MDragWidget.h"

const char* KChecked = ":/images/item-checked.png";
const char* KUnchecked = ":/images/item-unchecked.png";
const char* KHierarchy = ":/images/hierarchy-indicator.png";
const char* KShowThumbnail = ":/images/thumbnail-show.png";
const char* KWaitThumbnail = ":/images/thumbnail-wait.png";

CheckableListItem::CheckableListItem(CheckableListItem::TItemType aItemType,
                                     const QString& aItemId, bool aChecked, MDragWidget& aParent)
                                         : QFrame(&aParent), ui(new Ui::CheckableListItem), iItemType(aItemType), iItemId(aItemId),
                                         iChecked(aChecked), iParent(aParent), iThumbnailCreated(false)
{
    static const QPixmap showThumbnail(KShowThumbnail);
    ui->setupUi(this);

    if (iItemType == ECategory)
    {
        setObjectName("categoryListItem");
    	ui->thumbLabel->setEnabled(false);
    	ui->thumbLabel->hide();
    	ui->modifiedLabel->setEnabled(false);
    	ui->modifiedLabel->hide();
    	ui->sizeLabel->setEnabled(false);
    	ui->sizeLabel->hide();
    }
    else
    {
    	setObjectName("imageListItem");
    	ui->indentLabel->setEnabled(false);
    	ui->indentLabel->hide();
    	// Set an arrow that shows the thumbnail
    	ui->thumbLabel->setPixmap(showThumbnail);
    }
    
    UpdateCheckState();
}

CheckableListItem::~CheckableListItem()
{
    delete ui;
}

bool CheckableListItem::IsChecked() const
{
    return iChecked;
}

void CheckableListItem::SetCheckState(bool aState)
{
    iChecked = aState;
    UpdateCheckState();
}

void CheckableListItem::UpdateCheckState()
{
    // Style sheets according to the selected state and item type
    if (iChecked) {
        setStyleSheet("QFrame { background-color: #B2D9FB; }");
    } else {
        setStyleSheet("QFrame { background-color: white; }");
    }
}

void CheckableListItem::mouseReleaseEvent(QMouseEvent *e)
{
    static const QPixmap waitThumbnail(KWaitThumbnail);
    if (hasFocus() && !iParent.MDragWidgetIsDragging())
    {
        const QPoint position = ui->thumbLabel->mapTo(this, QPoint(0,0));
        if (iItemType == EImage && !iThumbnailCreated && QRect(position, ui->thumbLabel->size()).contains(e->pos()))
        {
            ui->thumbLabel->setPixmap(waitThumbnail);
            // Instead of a simple timer, we could use asynchronous image loader
            QTimer::singleShot(30, this, SLOT(CreateThumbnail()));
        }
        else
        {
            iChecked = !iChecked;
            UpdateCheckState();
        }
    }
    QFrame::mouseReleaseEvent(e);
}

void CheckableListItem::keyReleaseEvent(QKeyEvent *e)
{
    if (e->key() == Qt::Key_Select)
    {
        iChecked = !iChecked;
        UpdateCheckState();
    }
    else if (e->key() == Qt::Key_Right && iItemType == EImage && !iThumbnailCreated)
    {
        CreateThumbnail();
    }
}

void CheckableListItem::changeEvent(QEvent *e)
{
    QFrame::changeEvent(e);
    switch (e->type()) {
    case QEvent::LanguageChange:
        ui->retranslateUi(this);
        break;
    default:
        break;
    }
}

const QString& CheckableListItem::ItemId() const
{
    return iItemId;
}

void CheckableListItem::InitImageItem(const QString& aFilename, const QString& aDate, const QString& aSize)
{
    ui->titleLabel->setText(aFilename);
    ui->modifiedLabel->setText(aDate);
    ui->sizeLabel->setText(aSize);
}

void CheckableListItem::InitCategoryItem(const QString& aCategoryName, int aCategoryLevel)
{
    static const QPixmap iconHierarchy(KHierarchy);
    ui->titleLabel->setText(aCategoryName);
    // This can't be defined in .qss
    if (aCategoryLevel == 0) {
        ui->indentLabel->hide();
    } else {
        ui->indentLabel->setIndent(aCategoryLevel * 20);
        ui->indentLabel->setMinimumWidth(40 + (aCategoryLevel - 1) * 20);
        ui->indentLabel->setPixmap(iconHierarchy);
    }
}

void CheckableListItem::CreateThumbnail()
{
    if (iThumbnailCreated)
        return;
    QImageReader reader;
    reader.setFileName(iItemId);
    QSize imageSize = reader.size();

    imageSize.scale(ui->thumbLabel->size(), Qt::KeepAspectRatio);
    reader.setScaledSize(imageSize);
    QImage image(reader.read());
    // Set an arrow that shows the thumbnail
    ui->thumbLabel->setPixmap(QPixmap::fromImage(image));
    iThumbnailCreated = true;
}

bool CheckableListItem::ThumbnailCreated() const
{
    return iThumbnailCreated;
}

void CheckableListItem::CreateThumbnail(const QImage& aImage)
{
    ui->thumbLabel->setPixmap(QPixmap::fromImage(aImage));
}

void CheckableListItem::ThumbnailCreationStarted()
{
    iThumbnailCreated = true;
}
