
#include <QBoxLayout>
#include <QAction>
#include <QFileInfo>
#include <QPushButton>
#include <QScrollBar>
#include <qcoreevent.h>
#include <QKeyEvent>

#include <math.h>
#include "ImageSelectionList.h"
#include "ImageSearchThread.h"
#include "ImageThumbnailThread.h"
#include "wpscrollarea.h"
#include "CheckableListItem.h"


ImageSelectionList::ImageSelectionList(const QStringList& aInitialSelections, QWidget* aParent)
    : MDragWidget(aParent), iInitialSelections(aInitialSelections)
{
    setWindowFlags(Qt::Window);
#ifdef Q_WS_MAEMO_5
    setAttribute(Qt::WA_Maemo5StackedWindow);
#endif
    setWindowTitle(tr("Image selection"));

    QVBoxLayout* layout = new QVBoxLayout(this);
    layout->setSpacing(0);
    layout->setContentsMargins(0, 0, 0, 0);
    iScrollArea = new WPScrollArea(this);
    iScrollArea->setFocusPolicy(Qt::NoFocus);
    layout->addWidget(iScrollArea);
    connect(iScrollArea, SIGNAL(ScrollingStopped()), this, SLOT(UpdateThumbnails()));
    connect(iScrollArea->verticalScrollBar(), SIGNAL(sliderReleased()), this, SLOT(UpdateThumbnails()));

    QAction* leftSoftKey = new QAction(tr("Select"), this);
    connect(leftSoftKey, SIGNAL(triggered()), this, SLOT(ItemsSelected()));
    leftSoftKey->setSoftKeyRole(QAction::PositiveSoftKey);
    //addAction(leftSoftKey);
    QAction* rightSoftKey = new QAction(tr("Cancel"), this);
    connect(rightSoftKey, SIGNAL(triggered()), this, SLOT(CancelClicked()));
    rightSoftKey->setSoftKeyRole(QAction::NegativeSoftKey);
    //addAction(rightSoftKey);

    setLayout(layout);

    iThumbnailThread = new ImageThumbnailThread(QSize(80, 80), this); // TODO: Size
    connect(this, SIGNAL(CreateThumbnail(QString, int)), iThumbnailThread, SLOT(AddImage(QString, int)), Qt::QueuedConnection);
    connect(this, SIGNAL(StartThumbnailCreation()), iThumbnailThread, SLOT(start()), Qt::QueuedConnection);
    connect(iThumbnailThread, SIGNAL(ThumbnailCreated(int, const QImage&)), this, SLOT(ThumbnailReady(int, const QImage&)), Qt::BlockingQueuedConnection);
    
    // Finally add the data for WPCustomViewBasicData
    //
    iTitleText = tr("Images");
    iTabText = tr("Images");
    iHappyAction = leftSoftKey;
    iSadAction = rightSoftKey;
}

ImageSelectionList::~ImageSelectionList()
{
    if (iSearchThread)
    {
        if (iSearchThread->isRunning())
        {
            iSearchThread->Cancel();
            iSearchThread->wait();
        }
        delete iSearchThread;
    }
    if (iThumbnailThread)
    {
        if (iThumbnailThread->isRunning())
        {
            iThumbnailThread->Cancel();
            iThumbnailThread->wait();
        }
        delete iThumbnailThread;
    }
}

void ImageSelectionList::SearchImages()
{
    emit TaskStarted(EIdle);
    iSearchThread = new ImageSearchThread(this);
    qRegisterMetaType<QFileInfo>("QFileInfo");
    connect(iSearchThread, SIGNAL(ImageFound(const QFileInfo&)), this, SLOT(ImageFound(const QFileInfo&)));
    connect(iSearchThread, SIGNAL(finished()), this, SLOT(ThreadFinished()));
    iSearchThread->start();
}

void ImageSelectionList::ImageFound(const QFileInfo& aFileInfo)
{
    iFiles.append(new QFileInfo(aFileInfo));
}

void ImageSelectionList::CancelClicked()
{
    emit ImagesSelected(iInitialSelections);
}

void ImageSelectionList::ItemsSelected()
{
    QStringList filenames;
    const int count = iItems.count();
    for (int i = 0; i < count; i++)
    {
        if (iItems[i]->IsChecked())
            filenames.append(iItems[i]->ItemId());
    }
    emit ImagesSelected(filenames);
#ifdef Q_WS_MAEMO_5
    close();
#endif
}

void ImageSelectionList::ThreadFinished()
{
    if (!iScrollArea->widget())
        iScrollArea->InitLayout();
    iScrollArea->widget()->setMinimumHeight(iFiles.count() * 80);
    foreach (const QFileInfo* aFileInfo, iFiles)
    {
        CheckableListItem* item = new CheckableListItem(CheckableListItem::EImage,
                                                        aFileInfo->filePath(),
                                                        iInitialSelections.contains(aFileInfo->filePath()),
                                                        *this);
        item->setFixedHeight(80);
        item->InitImageItem(aFileInfo->fileName(),
                            WPNetworkEngine::LocalizedTime(aFileInfo->lastModified()),
                            FormatFileSize(aFileInfo->size()));

        iScrollArea->addWidget(item);
        iItems.append(item);
        // Used, so remember to delete
        delete aFileInfo;
    }

    emit TaskFinished(EIdle);

    iFiles.clear();
    delete iSearchThread;
    iSearchThread = NULL;
    iScrollArea->AddStretch();
    startTimer(1000);
}

bool ImageSelectionList::MDragWidgetIsDragging() const
{
    return iScrollArea->IsDragging();
}

QSize ImageSelectionList::VisibleSize() const
{
    return iScrollArea->GetVisibleContentSize();
}

QString ImageSelectionList::FormatFileSize(int aFileSize) const
{
    // TODO make something more fancy
    return QString::number(floor(aFileSize / 1024)) + tr(" kB");
}

void ImageSelectionList::keyReleaseEvent(QKeyEvent * event)
{
    //qDebug("WP::ImageSelectionList::keyReleaseEvent Key event %d", event->key());
    if (event->key() == Qt::Key_Context1 && iHappyAction)
    {
        iHappyAction->trigger();
        event->accept();
    }
    else if (event->key() == Qt::Key_Context2 && iSadAction)
    {
        iSadAction->trigger();
        event->accept();
    }
    else
    {
        event->ignore();
    }
}

void ImageSelectionList::setFocus()
{
    if (iItems.count() > 0)
        iItems.at(0)->setFocus();
    else
        iScrollArea->setFocus();
}

void ImageSelectionList::UpdateThumbnails()
{
    // TODO: Binary search to decrease from O(N) to O(log(N))
    int i = 0;
    while (i < iItems.size() &&
           iItems[i]->mapTo(iScrollArea->widget(), QPoint(0,0)).y() + iItems[i]->size().height() <
           -iScrollArea->widget()->pos().y())
    {
        i++;
    }
    while (i < iItems.size() && iItems[i]->mapTo(iScrollArea->widget(), QPoint(0,0)).y() <
           -iScrollArea->widget()->pos().y() + iScrollArea->viewport()->size().height())
    {
        if (!iItems[i]->ThumbnailCreated())
        {
            emit CreateThumbnail(iItems[i]->ItemId(), i);
            iItems[i]->ThumbnailCreationStarted();
        }
        i++;
    }
    emit StartThumbnailCreation();
}

void ImageSelectionList::ThumbnailReady(int aIndex, const QImage& aImage)
{
    iItems[aIndex]->CreateThumbnail(aImage);
}

void ImageSelectionList::timerEvent(QTimerEvent* aEvent)
{
    UpdateThumbnails();
    killTimer(aEvent->timerId());
}
