
#include <QGraphicsGridLayout>
#include <QGraphicsLinearLayout>
#include <QGraphicsProxyWidget>
#include <QGraphicsSceneMouseEvent>

#include <QApplication>
#include <QEvent>
#include <QPoint>
#include <QKeyEvent>
#include <QRect>
#include <QDesktopWidget>
#include <QPainter>

#include "clippinglabel.h"
#include "WPData.h"
#include "wpcommentgraphicslist.h"
#include "wpcommentgraphicslistitem.h"
#include "WPNetworkEngine.h"
#include "MDragWidget.h"
#include "themedatastorage.h"
#include "busyindicator.h"
#include "wpcommongeometry.h"

WPCommentGraphicsListItem::StandardGeometry WPCommentGraphicsListItem::iStandardGeometry;

WPCommentGraphicsListItem::WPCommentGraphicsListItem(const WPComment aComment, bool aShowFullComment, BusyIndicator& aBusyIndicator, ThemeData& aThemeData)
	: WPGraphicsListItem(aBusyIndicator, aThemeData)
    , iComment(aComment)
    , iShowFullComment(aShowFullComment)
{
    
    iDate = aComment->GetShortLocalTimeString(); 
}

WPCommentGraphicsListItem::~WPCommentGraphicsListItem()
{
}

void WPCommentGraphicsListItem::SetAvatarImage(const QPixmap &aPixmap)
{
    if (!aPixmap.isNull())
        iPixmap = aPixmap;
}

WPComment WPCommentGraphicsListItem::GetComment() const
{
    return iComment;
}

void WPCommentGraphicsListItem::Update(WPComment aComment)
{
    // XXX: no-op?
    iComment = aComment;
}

bool WPCommentGraphicsListItem::operator<(WPCommentGraphicsListItem const & other)
{
    return iComment->GetLocalTime() < other.iComment->GetLocalTime();
}

void WPCommentGraphicsListItem::ChangeViewMode()
{
    prepareGeometryChange();
    iShowFullComment = !iShowFullComment;
    SetViewMode(iShowFullComment);
}

void WPCommentGraphicsListItem::paintContent(QPainter* painter, const QStyleOptionGraphicsItem* option, QWidget* /*widget*/)
{    
    // Common part
    if (!iPixmap.isNull())
    {
        // TODO default avatar from themestorage?
        painter->drawPixmap(iStandardGeometry.iPixmap.topLeft(), iPixmap);
    }
    
    painter->setPen(Qt::black);
    painter->setFont(*iThemeData.iListDetailFont);
    painter->drawText(iStandardGeometry.iAuthor, iComment->iAuthor);
    painter->drawText(iStandardGeometry.iDate, iDate);
    
    if (iShowFullComment)
    {
        painter->setFont(*iThemeData.iListInfoFont);
        painter->drawText(iExpandedGeometry.iTitle, iComment->iPostTitle);
        
        painter->setFont(*iThemeData.iListContentFont);
        painter->drawText(iExpandedGeometry.iComment, iComment->iContent);
        
        // XXX elide
        painter->drawText(iExpandedGeometry.iEmail, iComment->iAuthorEmail);
        painter->drawText(iExpandedGeometry.iWebsite, iComment->iAuthorUrl);
    }
    else
    {
        painter->setFont(*iThemeData.iListInfoFont);
        painter->drawText(iStandardGeometry.iTitle, iTitle);
        
        painter->setFont(*iThemeData.iListContentFont);
        painter->drawText(iStandardGeometry.iComment, iCommentText);
        
        // XXX elide
        painter->drawText(iStandardGeometry.iEmail, iComment->iAuthorEmail);
        painter->drawText(iStandardGeometry.iWebsite, iComment->iAuthorUrl);
    }
}

bool WPCommentGraphicsListItem::IsBusy()
{
    return iComment->IsBusy();
}

bool WPCommentGraphicsListItem::HasError()
{
    return iComment->iError;
}

void WPCommentGraphicsListItem::handleGeomeryChange()
{
	SetViewMode(iShowFullComment);
}

void WPCommentGraphicsListItem::calculateElidedTexts()
{
    if(!iShowFullComment)
    {
        QFontMetrics titleFm(*iThemeData.iListInfoFont);
        iTitle = titleFm.elidedText(iComment->iPostTitle, Qt::ElideRight, iStandardGeometry.iTitle.width());
        QFontMetrics commentFm(*iThemeData.iListContentFont);
        iCommentText = commentFm.elidedText(iComment->iContent, Qt::ElideRight, iStandardGeometry.iComment.width());
    }
    
    // XXX elide email and website
}

void WPCommentGraphicsListItem::SetViewMode(bool aShowFullComment)
{
    if (aShowFullComment)
    {
        iExpandedGeometry.update(*this);
        setMinimumHeight(iExpandedGeometry.height());
    }
    else
    {
        iStandardGeometry.update(size(), iThemeData);
        setMinimumHeight(iStandardGeometry.height());
    }
    
    calculateElidedTexts();
}

WPCommentGraphicsListItem::StandardGeometry::StandardGeometry()
  : iHeight(0)
    { }

void WPCommentGraphicsListItem::StandardGeometry::update(QSizeF const & size, ThemeData const & theme)
    {
    if(iInitSize == size) { return; }
    
    // XXX
    int gravatarSize = 80;
    int padding = theme.iListItemPadding;
    int edgePadding = padding + theme.iListItemShadowMargin;
    
    // Font metrics
    QFontMetrics detailFm(*theme.iListDetailFont);
    QFontMetrics infoFm(*theme.iListInfoFont);
    QFontMetrics contentFm(*theme.iListContentFont);
    
    // Layout rects
    iPixmap.setRect(edgePadding, edgePadding, gravatarSize, gravatarSize);
    iAuthor.setRect(edgePadding, iPixmap.bottom() + padding, gravatarSize, detailFm.height());
    iDate.setRect(edgePadding, iAuthor.bottom() + padding, gravatarSize, detailFm.height());
    
    int remainingWidth = size.width() - iPixmap.right() - (padding + edgePadding);
    int col2x = iPixmap.right() + padding;
    iTitle.setRect(col2x, edgePadding, remainingWidth, infoFm.height() );
    iComment.setRect(col2x, iTitle.bottom() + 3 * padding, remainingWidth, contentFm.height());
    iEmail.setRect(col2x, iComment.bottom() + 3 * padding, remainingWidth, detailFm.height());
    iWebsite.setRect(col2x, iEmail.bottom(), remainingWidth, detailFm.height());
    
    updateHeight(theme);
    iInitSize = size;
    }

void WPCommentGraphicsListItem::StandardGeometry::updateHeight(ThemeData const & theme)
    {
    iHeight = std::max(iDate.bottom(), iWebsite.bottom()) +
              theme.iListItemPadding + theme.iListItemShadowMargin;
    }

void WPCommentGraphicsListItem::ExpandedGeometry::update(WPCommentGraphicsListItem & parent)
    {
    if(iInitSize == parent.size()) { return; }
    
    StandardGeometry::update(parent.size(), parent.iThemeData);

    QFontMetrics infoFm(*parent.iThemeData.iListInfoFont);
    QFontMetrics contentFm(*parent.iThemeData.iListContentFont);
    
    int padding = parent.iThemeData.iListItemPadding;
    const int maxHeight(1000);
    
    iTitle.setHeight(maxHeight);
    iTitle = infoFm.boundingRect(iTitle, Qt::TextWordWrap, parent.iComment->iPostTitle);
    
    iComment.setHeight(maxHeight);
    iComment = contentFm.boundingRect(iComment, Qt::TextWordWrap, parent.iComment->iContent);
    iComment.moveTop(iTitle.bottom() + 3 * padding);
    
    iEmail.moveTop(iComment.bottom() + 3 * padding);
    iWebsite.moveTop(iEmail.bottom());
    
    updateHeight(parent.iThemeData);
    iInitSize = parent.size();
    }

