#include "wpgraphicslist.h"

#include <QGraphicsLinearLayout>
#include <QGraphicsView>
#include <QGraphicsSceneResizeEvent>
#include <QPropertyAnimation>

#include "wpgraphicslistitem.h"
#include "wpscrollarea.h"
#include "PopupMenu.h"
#include "themedatastorage.h"
#include "enume/pannableview.h"
#include "BlogView.h"
#include "wpkeydispatcher.h"
#include "wpcommongeometry.h"
#include "utils.h"

WPGraphicsList::WPGraphicsList(BlogView* aBlogView, const QGraphicsView& aGraphicsView,
	        PannableView* aPannableView)
  : focusFunctor(*this)
  , iCurrentItem(0, focusFunctor)
  , iBusyIndicator(aBlogView->GetBusyIndicator())
  , iThemeData(aBlogView->GetThemeData())
  , iGraphicsView(aGraphicsView) 
  , iKeyDispatcher(aBlogView)
  , iPopupMenu(NULL)
#ifndef Q_WS_MAEMO_5
  , iAnimation(NULL)
  , iMenuHiding(false)
#endif /* Q_WS_MAEMO_5 */
  , iPannableView(aPannableView)
{
	iLayout = new QGraphicsLinearLayout(Qt::Vertical);
	iLayout->setSpacing(iThemeData.iListItemSpacing);
	iLayout->setContentsMargins(5, 0, 5, 0);
    setContentsMargins(0, 0, 0, 0);
    //iLayout->addStretch();
    setLayout(iLayout);
}

QGraphicsLinearLayout * WPGraphicsList::AddLayout()
{
    QGraphicsLinearLayout * childLayout = new QGraphicsLinearLayout(Qt::Vertical);
    childLayout->setSpacing(iThemeData.iListItemSpacing);
    childLayout->setContentsMargins(0, 0, 0, 0);
    iLayout->addItem(childLayout);
    childLayout->setParentLayoutItem(iLayout);
    return childLayout;
}

void WPGraphicsList::ClearLayout(QGraphicsLinearLayout & layout)
{
    while(layout.count()) {
        QGraphicsItem * item = layout.itemAt(0)->graphicsItem();
        if(iCurrentItem == item) { iCurrentItem = 0; }
        layout.removeAt(0);
        delete item;
    }
}

void WPGraphicsList::CollapseItem(QGraphicsLayoutItem * item)
{
    item->setMaximumSize(0, 0);
}

void WPGraphicsList::UncollapseItem(QGraphicsLayoutItem * item)
{
    item->setMaximumSize(INT_MAX, INT_MAX);
}

void WPGraphicsList::ShowContextMenu(WPGraphicsListItem* aItem)
{
    iCurrentItem = aItem;
    WPGraphicsListItem * listItem = CurrentListItem();
    if(!listItem) { return; }
    
    if (iPopupMenu)
        return;

    if (aItem->IsBusy() || aItem->BogusFocus())
        return;

    iPopupMenu = new PopupMenu;
    connect(iPopupMenu, SIGNAL(aboutToHide()), this, SLOT(MenuHiding()));
    iCurrentItem = aItem;
    listItem->MarkBogusFocus(true);
    
    PopulatePopupMenuForCurrentItem(iPopupMenu);
    
    iPopupMenu->setWindowOpacity(0.6);
    iPopupMenu->show();

    //QPoint itemPoint = iGraphicsView.mapFromScene(aItem->TopRight());
    //itemPoint = getMenuLocation(itemPoint);
    QPoint viewPoint = iGraphicsView.mapFromScene(aItem->scenePos());
    
    int x = viewPoint.x() + iGraphicsView.viewport()->width();
    int y = viewPoint.y();
    
    // fix out of screen
    //
    if (y + iPopupMenu->height() > iGraphicsView.viewport()->height())
    {
        y -= y + iPopupMenu->height() - iGraphicsView.viewport()->height();
    }
    if (y < 0) { y = 0; }

    // destination point to place popup menu at
    QRect dest(x - iPopupMenu->width(), y, iPopupMenu->width(), iPopupMenu->height());

// no animations on maemo    
#ifdef Q_WS_MAEMO_5
    iPopupMenu->setGeometry(dest);

#else    
    if (iAnimation)
    {
        delete iAnimation;
        iAnimation = NULL;
    }

    iAnimation = new QPropertyAnimation(iPopupMenu, "geometry", this);
    iAnimation->setDuration(200);

    qDebug("WP::Animating iPopupMenu FROM %d, %d TO %d, %d", x, y, x - iPopupMenu->width(), y);

    iAnimation->setStartValue(dest.adjusted(iPopupMenu->width(), 0, iPopupMenu->width(), 0));
    iAnimation->setEndValue(dest);
    iAnimation->setEasingCurve(QEasingCurve::Linear);
    iAnimation->start();

#endif /* Q_WS_MAEMO_5 */
}

void WPGraphicsList::ShowContextMenuForCurrent()
{
    WPGraphicsListItem * listItem = CurrentListItem();
    if(listItem)
        ShowContextMenu(listItem);
}

void WPGraphicsList::ContextMenuHidden()
{
#ifndef Q_WS_MAEMO_5
    iMenuHiding = false;
    delete iAnimation;
    iAnimation = NULL;
#endif

    iPopupMenu->deleteLater();
    iPopupMenu = NULL;

    setFocus();
}

void WPGraphicsList::MenuHiding()
{
#ifdef Q_WS_MAEMO_5
    // direct
    ContextMenuHidden();

#else    
    // never take this event twice.
    qDebug("WP::WPGraphicsList::MenuHiding()");
    if (iMenuHiding) return;
    
    if (iAnimation)
    {
        delete iAnimation;
        iAnimation = NULL;
    }

    if(!iPopupMenu) { return; }
    
    iPopupMenu->show();

    iAnimation = new QPropertyAnimation(iPopupMenu, "geometry", this);
    iAnimation->setDuration(200);
    
    QRectF itemGeometry = iPopupMenu->geometry();
    
    int x = int(itemGeometry.x());
    int y = int(itemGeometry.y());

    qDebug("WP::Animating iPopupMenu FROM %d, %d TO %d, %d", x, y, x + iPopupMenu->width(), y);

    iAnimation->setStartValue(QRect(x,y,iPopupMenu->width(),iPopupMenu->height()));
    iAnimation->setEndValue(QRect(x+iPopupMenu->width(),y,iPopupMenu->width(),iPopupMenu->height()));
    iAnimation->setEasingCurve(QEasingCurve::Linear);
    iAnimation->start();

    iMenuHiding = true;
    connect(iAnimation,SIGNAL(finished()), this, SLOT(ContextMenuHidden()));
#endif    
}

void WPGraphicsList::setFocus()
{
    WPGraphicsListItem * listItem = CurrentListItem();
    if (listItem)
    {
        listItem->setFocus();
        listItem->MarkBogusFocus(false); // gained real focus, disable bogus
    }
    
    if(iCurrentItem)
    {
        Utils::ForceFocus(iCurrentItem);
    }
    else if(iLayout->count())
    {
        // Focus first focusable item
        iCurrentItem = iLayout->itemAt(0);
        /// XXX fix iterator so we don't need this
        ++iCurrentItem;
        --iCurrentItem;
        Utils::ForceFocus(iCurrentItem);
    }
    
    QGraphicsWidget::setFocus();
}

void WPGraphicsList::resizeEvent(QGraphicsSceneResizeEvent* event)
{
    QGraphicsWidget::resizeEvent(event);
    
    WPGraphicsListItem * listItem = CurrentListItem();
    if (iPopupMenu && iPopupMenu->isVisible() && listItem)
        iPopupMenu->move(getMenuLocation(listItem->TopRight().toPoint()));
}

QSizeF WPGraphicsList::sizeHint(Qt::SizeHint which, const QSizeF & constraint) const
    {
    return iLayout->sizeHint(which, constraint);
    }

QPoint WPGraphicsList::getMenuLocation(QPoint itemPoint)
{
    const int xCorrection = itemPoint.x() + iPopupMenu->sizeHint().width() - iGraphicsView.viewport()->size().width();
    itemPoint += QPoint(-xCorrection, 0);
    return itemPoint;
}

void WPGraphicsList::keyPressEvent ( QKeyEvent * event  )
{
    qDebug("WP::WPGraphicsList::keyPressEvent Key event %d", event->key());
    
    if (event->key() == Qt::Key_Down)
    {
        ++iCurrentItem;
    }
    else if (event->key() == Qt::Key_Up)
    {
        --iCurrentItem;
    }
    else
    {
        return;
    }
        
    if(CurrentListItem())
        {
        iPannableView->moveToItem(iCurrentItem);
        setFocus();
        }
    else
        {
        Utils::ForceFocus(iCurrentItem);
        iPannableView->moveToItem(iCurrentItem);
        }
    
    event->accept();
}

void WPGraphicsList::keyReleaseEvent ( QKeyEvent * event  )
{
    if (event->key() != Qt::Key_Down &&
        event->key() != Qt::Key_Up )
    {
        iKeyDispatcher->keyClickedEvent(event);
    }
}

WPGraphicsListItem * WPGraphicsList::CurrentListItem()
{
    if(iCurrentItem) {
        return qgraphicsitem_cast<WPGraphicsListItem *>(iCurrentItem);
    }
    return 0;
}

bool WPGraphicsList::IsListItem(QGraphicsItem * item)
{
    return (qgraphicsitem_cast<WPGraphicsListItem *>(item) != 0);
}


WPGraphicsList::IteratorGuard::IteratorGuard(WPGraphicsList & list, QGraphicsItem * item)
  : list(list)
  , decremented(false)
{
    if(list.iCurrentItem == item) {
        if(!--list.iCurrentItem) {
            list.iCurrentItem  = 0;
        }
        decremented = true;
    }
}

WPGraphicsList::IteratorGuard::~IteratorGuard()
{
    if(list.iCurrentItem && decremented) {
        ++list.iCurrentItem;
    }
    list.setFocus();
}
