
#include <QGraphicsLinearLayout>
#include <QGraphicsView>
#include <QGraphicsSceneResizeEvent>
#include <QLabel>
#include <QPropertyAnimation>

#include "wppostgraphicslistitem.h"
#include "wppostgraphicslist.h"
#include "WPData.h"
#include "WPNetworkEngine.h"
#include "wpscrollarea.h"
#include "BlogView.h"
#include "PopupMenu.h"
#include "themedatastorage.h"
#include "enume/pannableview.h"
#include "wpkeydispatcher.h"
#include "wpconfirmdialog.h"

#include "buttongraphicsproxy.h"

WPPostGraphicsList::WPPostGraphicsList(
        const QGraphicsView& aGraphicsView,BlogView* aBlogView, PannableView* aPannableView
    ) : 
		WPGraphicsList(aBlogView, aGraphicsView, aPannableView)
{   
    // TODO find a proper way to align menu icon and text side by side.
    iActionOpen = new QAction(iThemeData.iContextMenuIcons[ThemeData::EOpen],tr("    Open"), this);
    iActionEdit = new QAction(iThemeData.iContextMenuIcons[ThemeData::EEdit],tr("    Edit"), this);
    iActionDelete = new QAction(iThemeData.iContextMenuIcons[ThemeData::ERemove],tr("    Delete"), this);

    connect(iActionOpen, SIGNAL(triggered(bool)), this, SLOT(Open()));
    connect(iActionEdit, SIGNAL(triggered(bool)), this, SLOT(Edit()));
    connect(iActionDelete, SIGNAL(triggered(bool)), this, SLOT(Delete()));

    iNewPostButton = new ButtonGraphicsProxy(tr("Write new post"), this);
    iNewPostButton->button()->setObjectName(QString::fromLatin1("newPostLabel"));
    //iNewPostButton->setFixedHeight(KTitleDefaultHeight);
    //iNewPostButton->setStyleSheet(QLatin1String(KButtonStyleSheet));
    connect(iNewPostButton->button(), SIGNAL(clicked(bool)), aBlogView, SLOT(ShowWriteNewPostView()));

    // labels, not automatically added 
    iDraftsLabel = new QLabel;
    iDraftsProxy = new QGraphicsProxyWidget(this);
    iDraftsProxy->setWidget(iDraftsLabel);
    iDraftsLabel->setObjectName(QString::fromLatin1("draftsLabel"));
    iDraftsLabel->setFixedHeight(iThemeData.iListTitleHeight);

    iPostsLabel = new QLabel;
    iPostsProxy = new QGraphicsProxyWidget(this);
    iPostsProxy->setWidget(iPostsLabel);
    iPostsLabel->setObjectName(QString::fromLatin1("postsLabel"));
    iPostsLabel->setFixedHeight(iThemeData.iListTitleHeight);
    
    iLayout->addItem(iNewPostButton);
    iLayout->addItem(iDraftsProxy);
    iDraftsLayout = AddLayout();
    iLayout->addItem(iPostsProxy);
    iPostsLayout = AddLayout();
    
    UpdateLabels();
}

WPPostGraphicsList::~WPPostGraphicsList()
{

}


void WPPostGraphicsList::PopulatePopupMenuForCurrentItem(PopupMenu* aPopupMenu)
{
    aPopupMenu->addAction(iActionOpen);
    aPopupMenu->addAction(iActionEdit);
    aPopupMenu->addAction(iActionDelete);
}

void WPPostGraphicsList::AddPost(WPPost aPost)
{
    QStringList list;

    if (aPost->IsLocal())
        list << "Local draft";

    list << WPNetworkEngine::LocalizedTime(aPost->iDateCreated);

    WPPostGraphicsListItem* item = new WPPostGraphicsListItem(aPost, iBusyIndicator, iThemeData);
    connect(item, SIGNAL(Selected(WPGraphicsListItem*)), this, SLOT(ShowContextMenu(WPGraphicsListItem*)));

    if(aPost->IsDraft()) {
        OrderedInsert(*iDraftsLayout, *item);
    } else {
        OrderedInsert(*iPostsLayout, *item);
    }
}

void WPPostGraphicsList::RemovePost (WPPost aPost)
{
    WPPostGraphicsListItem * item = 0;
        
    item = RemoveFromLayout(*iDraftsLayout, aPost);
    if(!item) { item = RemoveFromLayout(*iPostsLayout, aPost); }
    
    delete item;
    UpdateLabels();
}

void WPPostGraphicsList::Open()
{
    qDebug("WP::WPPostGraphicsList::Open");
    emit OpenClicked(CurrentItem()->GetPost());
}

void WPPostGraphicsList::Edit()
{
    qDebug("WP::WPPostGraphicsList::Edit");
    emit EditClicked(CurrentItem()->GetPost());
}

void WPPostGraphicsList::Delete()
{
    WP_CONFRIM_AND_RETURN_ON_FALSE

    qDebug("WP::WPPostGraphicsList::Delete");
    emit DeleteClicked(CurrentItem()->GetPost());
}

void WPPostGraphicsList::UpdatePostData(const WPDataBlog& aBlog)
{
    if(iCurrentBlog) {
        disconnect(iCurrentBlog, NULL, this, NULL);
    }
    iCurrentBlog = const_cast<WPDataBlog *>(&aBlog);
    
    ClearLayout(*iDraftsLayout);
    ClearLayout(*iPostsLayout);
    
    QListIterator<WPPost> posts(aBlog.GetPosts());
    while (posts.hasNext()) {
        WPPost post = posts.next();
        AddPost(post);
    }
    
    connect(
       &aBlog,  SIGNAL(PostAdded(WPPost)), 
       this,    SLOT(PostAdded(WPPost))
    );
    connect(
       &aBlog,  SIGNAL(PostChanged(WPPost)), 
       this,    SLOT(PostChanged(WPPost))
    );
    connect(
       &aBlog,  SIGNAL(PostRemoved(WPPost)), 
       this,    SLOT(PostRemoved(WPPost))
    );
       
    UpdateLabels();
}

void WPPostGraphicsList::PostChanged(WPPost aPost)
{
    qDebug("WPPostGraphicsList::PostChanged");

    // cycle
    RemovePost(aPost);
    AddPost(aPost);
    
    UpdateLabels();
}

void WPPostGraphicsList::PostRemoved(WPPost aPost)
{
    qDebug("WPPostGraphicsList::PostRemoved");

    RemovePost(aPost);
    UpdateLabels();
}

void WPPostGraphicsList::PostAdded(WPPost aPost)
{
    qDebug("WPPostGraphicsList::PostAdded");

    AddPost(aPost);
    UpdateLabels();
}

void WPPostGraphicsList::UpdateLabels()
{
    int drafts = iDraftsLayout->count();
    if (drafts) {
        iDraftsLabel->setText(tr("Drafts") + " (" + QString::number(drafts) + ')');
        UncollapseItem(iDraftsProxy);
    } else {
        CollapseItem(iDraftsProxy);
    }

    int posts = iPostsLayout->count();
    if (posts) {
        iPostsLabel->setText(tr("Posts") + " (" + QString::number(posts) + ')');
        UncollapseItem(iPostsProxy);
    } else {
        CollapseItem(iPostsProxy);
    }
}

void WPPostGraphicsList::OrderedInsert(QGraphicsLinearLayout & layout, WPPostGraphicsListItem & item)
{
    int i;
    for(i = 0; i < layout.count(); ++i) {
        WPPostGraphicsListItem * listItem = LayoutToListItem(layout.itemAt(i));
        if(*listItem < item) { break; }
    }
    layout.insertItem(i, &item);
}

WPPostGraphicsListItem * WPPostGraphicsList::RemoveFromLayout(QGraphicsLinearLayout & layout, WPPost post)
{
    for(int i = 0; i < layout.count(); ++i) {
        WPPostGraphicsListItem * item = LayoutToListItem(layout.itemAt(i));
        if(item->GetPost() == post) {
            IteratorGuard itGuard(*this, item);
            layout.removeAt(i);
            return item;
        }
    }
    return 0;
}

bool WPPostGraphicsList::IsFocusable(QGraphicsLayoutItem * item)
{
    QGraphicsItem * gItem = item->graphicsItem();
    return (WPGraphicsList::IsListItem(gItem) || gItem == iNewPostButton); 
}

